/*
	tcfi_conv	TCFI Converter version 1.4 - June 17. 2000

	Copyright (c) by Collin R. Mulliner <palm@mulliner.org>
	WebSite: www.mulliner.org/palm

	for more information read: readme.txt
*/

#include <stdio.h>
#include <stdlib.h>

#define MAX_REC_SIZE 0xff00

#define HELP_TEXT	"TCFI Converter (commandline version) 1.4\n"\
			"Copyright (c) by Collin R. Mulliner <palm@mulliner.org>\n"\
			"WebSite: www.mulliner.org/palm\n\n"\
			"Syntax: <file name>.XXX <file name>.pdb\n"

typedef struct {
	char file_name[256];

	unsigned long file_size;
} tcfi_db_info_block_type;

void write_data(FILE *in, FILE *out, unsigned long file_size, tcfi_db_info_block_type *tcfi_block);
unsigned long get_file_size(FILE *fp);
void build_tcfi_db_info_block(tcfi_db_info_block_type *info_block, char *name, FILE *fp);
void write_pdb_header(FILE *fp, char name[], int num_records);
void write_record_headers(FILE *fp, int num_records, unsigned long file_size);
void make_offset(unsigned long to, unsigned char *oo);

int main(int argc, char **argv)
{
	FILE *fp, *pdb;
	tcfi_db_info_block_type file_info;
	unsigned long num_recs, tmp;
	unsigned long orig_file_size;

	
	if (argc == 3) {
		fp = fopen(argv[1], "rb");
		build_tcfi_db_info_block(&file_info, argv[1], fp);

		pdb = fopen(argv[2], "wb");

		orig_file_size = get_file_size(fp);
		num_recs = orig_file_size;

		tmp = num_recs / MAX_REC_SIZE;
		if ((num_recs % MAX_REC_SIZE) > 0) num_recs = tmp + 1;
		else { num_recs = tmp; }
		num_recs++;

		#ifdef DEBUG
		printf("Num Records (inc. info_block rec): %d\n", num_recs);
		#endif

		write_pdb_header(pdb, file_info.file_name, num_recs);
		write_record_headers(pdb, num_recs - 1, orig_file_size);

		fclose(fp);
		fp = fopen(argv[1], "rb");
		// close and open source file - bad stuff

		write_data(fp, pdb, orig_file_size, &file_info);

		fclose(fp);
		fclose(pdb);
	}
	else {
		printf(HELP_TEXT);
	}
}

void write_data(FILE *in, FILE *out, unsigned long file_size, tcfi_db_info_block_type *tcfi_block)
{
	unsigned char *in_buf;


	fwrite(tcfi_block, sizeof(tcfi_db_info_block_type), 1, out);
	// write tcfi record

	in_buf = (unsigned char *) malloc(file_size);

	fread(in_buf, file_size, 1, in);
	fwrite(in_buf, file_size, 1, out);

	free(in_buf);
}

unsigned long get_file_size(FILE *fp)
{
	fseek(fp, 0, SEEK_END);
	return(ftell(fp));
}

void build_tcfi_db_info_block(tcfi_db_info_block_type *info_block, char *name, FILE *fp)
{
	unsigned long tmp_size;
	long i;
	

	i = strlen(name);
	do {
		i--;	
	} while (name[i] != '/' && name[i] != '\\' && i > 0);
	if (i == 0) i = -1;

	strncpy(info_block->file_name, name+i+1, (((strlen(name) - i) > 255) ? 255 : strlen(name) - i));
	// max file name length is 255 chars
	info_block->file_size = get_file_size(fp);
	make_offset(info_block->file_size, &tmp_size);
	info_block->file_size = tmp_size;
}

void write_pdb_header(FILE *fp, char name[], int num_records)
{
	int i;
	char db_name[33];

	for (i = 0; i < 32; i++) { db_name[i] = '\0'; }

	strncpy(db_name, name, 32);

	for (i = 0; i < 32; i++) {
		fputc(db_name[i], fp);
	}

	fputc(0x00, fp);
	fputc(0x00, fp);
	//	the flags
	fputc(0x00, fp);
	fputc(0x00, fp);
	//	the version 
	fputc(0x06, fp);
	fputc(0xD1, fp);
	fputc(0x44, fp);
	fputc(0xAE, fp);
	//	the creation time 
	fputc(0x06, fp);
	fputc(0xD1, fp);
	fputc(0x44, fp);
	fputc(0xAE, fp);
	//	the modifi time 
	fputc(0x00, fp);
	fputc(0x00, fp);
	fputc(0x00, fp);
	fputc(0x00, fp);
	//	the backup time 
	fputc(0x00, fp);
	fputc(0x00, fp);
	fputc(0x00, fp);
	fputc(0x00, fp);
	//	the modifi number 
	fputc(0x00, fp);
	fputc(0x00, fp);
	fputc(0x00, fp);
	fputc(0x00, fp);
	//	the app info offset 
	fputc(0x00, fp);
	fputc(0x00, fp);
	fputc(0x00, fp);
	fputc(0x00, fp);
	//	the sort info offset 

	fputc('T', fp);
	fputc('C', fp);
	fputc('F', fp);
	fputc('I', fp);
	//	the DB Type "TCFI" 
	fputc('C', fp);
	fputc('R', fp);
	fputc('M', fp);
	fputc('G', fp);
	//	the CRID 

	fputc(0x00, fp);
	fputc(0x00, fp);
	fputc(0x00, fp);
	fputc(0x00, fp);
	//	the unique ID seed 
	fputc(0x00, fp);
	fputc(0x00, fp);
	fputc(0x00, fp);
	fputc(0x00, fp);
	//	the next record list ID

	fputc(0x00, fp);
	fputc(num_records, fp);
 	// number of records
}

void write_record_headers(FILE *fp, int num_records, unsigned long file_size)
{
	unsigned long cur_rec_size;
	unsigned long new_offset;
	unsigned char offset_out[4];
	int i, i2;
	int cur_rec_no;


	cur_rec_no = 0;

	new_offset = 78 + ((num_records + 1) * 8);

	make_offset(new_offset, offset_out);

	for (i = 0; i < 4; i++) {
		fputc(offset_out[i], fp);
	}
	
	fputc(0x40, fp);
	fputc(0x6F, fp);
	fputc(0x80, fp);
	fputc(0x00, fp);
	
	cur_rec_size = sizeof(tcfi_db_info_block_type);

	cur_rec_no = 1;
	for (i = cur_rec_no; i < num_records + 1; i++) {
		new_offset = new_offset + cur_rec_size;

		make_offset(new_offset, offset_out);

		for (i2 = 0; i2 < 4; i2++) {
			fputc(offset_out[i2], fp);
		}
	
		fputc(0x40, fp);
		fputc(0x6F, fp);
		fputc(0x80, fp);
		fputc((unsigned char) i, fp);

		cur_rec_size = ((file_size >= MAX_REC_SIZE) ? MAX_REC_SIZE : file_size);
		file_size -= cur_rec_size;
	}
}

void make_offset(unsigned long to, unsigned char *oo)
{
	unsigned long z3, zcc, zcc2, zcc3;

	z3 = to;
	zcc3 = z3 / (65535 * 256);
	z3 = z3 - (zcc3 * (65535 * 256));
	zcc2 = z3 / 65535;
	z3 = z3 - (zcc2 * 65535);
	zcc = z3 / 256;
	*(oo+3) = z3 - (zcc * 256);
	*(oo+2) = zcc;
	*(oo+1) = zcc2;
	*(oo) = zcc3;
}
